/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.extension.index.security;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.xwiki.extension.version.Version;
import org.xwiki.stability.Unstable;
import org.xwiki.text.XWikiToStringBuilder;

import us.springett.cvss.Cvss;

/**
 * An individual security vulnerability descriptor.
 *
 * @version $Id: d0ffe3d8d07bc463bbda0051b9f5d893c0375ba0 $
 * @since 15.5RC1
 */
@Unstable
public class SecurityVulnerabilityDescriptor
{
    private String id;

    private String url;

    private double score;

    private Version fixVersion;

    /**
     * @param id the security vulnerability id
     * @return the current object
     */
    public SecurityVulnerabilityDescriptor setId(String id)
    {
        this.id = id;
        return this;
    }

    /**
     * @return the security vulnerability id
     */
    public String getId()
    {
        return this.id;
    }

    /**
     * @param url an external URL providing more details on the vulnerability
     * @return the current object
     */
    public SecurityVulnerabilityDescriptor setURL(String url)
    {
        this.url = url;
        return this;
    }

    /**
     * @return an external URL providing more details on the vulnerability
     */
    public String getURL()
    {
        return this.url;
    }

    /**
     * Compute and store the score from the provided CVSS vector.
     *
     * @param vector a CVSS vector to parse and compute the based score from
     * @return the current object
     * @see #getScore()
     */
    public SecurityVulnerabilityDescriptor setSeverityScore(String vector)
    {
        this.score = Cvss.fromVector(vector).calculateScore().getBaseScore();
        return this;
    }

    /**
     * @return the CVSS score of the security vulnerability
     */
    public double getScore()
    {
        return this.score;
    }

    /**
     * @param score the CVSS score of the security vulnerability
     * @return the current object
     */
    public SecurityVulnerabilityDescriptor setScore(double score)
    {
        this.score = score;
        return this;
    }

    /**
     * @return the minimal version to which to upgrade to get the issue fixed automatically
     */
    public Version getFixVersion()
    {
        return this.fixVersion;
    }

    /**
     * @param fixVersion the minimal version to which to upgrade to get the issue fixed automatically
     * @return the current object
     */
    public SecurityVulnerabilityDescriptor setFixVersion(Version fixVersion)
    {
        this.fixVersion = fixVersion;
        return this;
    }

    @Override
    public boolean equals(Object o)
    {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        SecurityVulnerabilityDescriptor that = (SecurityVulnerabilityDescriptor) o;

        return new EqualsBuilder()
            .append(this.score, that.score)
            .append(this.id, that.id)
            .append(this.url, that.url)
            .append(this.fixVersion, that.fixVersion)
            .isEquals();
    }

    @Override
    public int hashCode()
    {
        return new HashCodeBuilder(17, 37)
            .append(this.id)
            .append(this.url)
            .append(this.score)
            .append(this.fixVersion)
            .toHashCode();
    }

    @Override
    public String toString()
    {
        return new XWikiToStringBuilder(this)
            .append("id", this.id)
            .append("url", this.url)
            .append("score", this.score)
            .append("fixVersion", this.fixVersion)
            .toString();
    }
}
