/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.extension.security.internal.analyzer;

import java.io.IOException;
import java.util.List;

import javax.inject.Inject;
import javax.inject.Singleton;

import org.apache.solr.client.solrj.SolrServerException;
import org.slf4j.Logger;
import org.xwiki.component.annotation.Component;
import org.xwiki.extension.Extension;
import org.xwiki.extension.index.internal.ExtensionIndexStore;
import org.xwiki.extension.index.security.ExtensionSecurityAnalysisResult;

import static org.apache.commons.lang3.exception.ExceptionUtils.getRootCauseMessage;

/**
 * Update an {@link Extension} with the results of its latest {@link ExtensionSecurityAnalysisResult}.
 *
 * @version $Id: 34b3356126b977516a68af22f23afb35f8059561 $
 * @since 15.5RC1
 */
@Component(roles = VulnerabilityIndexer.class)
@Singleton
public class VulnerabilityIndexer
{
    @Inject
    private ExtensionIndexStore extensionIndexStore;

    @Inject
    private Logger logger;

    /**
     * Update the security vulnerabilities known for the given extension.
     *
     * @param extension the extension to update
     * @param analysis the security analysis to update the extension
     * @return {@code true} if some new security vulnerabilities are inserted, {@code false} otherwise
     */
    public boolean update(Extension extension, ExtensionSecurityAnalysisResult analysis)
    {
        try {
            List<String> cveiDs = this.extensionIndexStore.getCVEIDs(extension.getId());

            boolean hasNew =
                analysis.getSecurityVulnerabilities().stream()
                    .anyMatch(vulnerability -> !cveiDs.contains(vulnerability.getId()));

            this.extensionIndexStore.update(extension.getId(), analysis);

            return hasNew;
        } catch (SolrServerException | IOException e) {
            this.logger.warn(
                "Failed to update the extension [{}] with the results of its latest security scan. Cause: [{}]",
                extension.getId(), getRootCauseMessage(e));
            return false;
        }
    }
}
